// =============================================================================
//	CleanupPlugin.c
// =============================================================================

#define DEBUGSTR(s) DebugStr(s)

// =============================================================================

#if defined(__MWERKS__)
#include <CFPlugInCOM.h>
#include <ApplicationServices.h>
#include <Carbon.h>
#include <CoreServices.h>
#include <Menus.h>

// The Contextual Menu Manager will only load CFPlugIns of type kContextualMenuTypeID
#define kContextualMenuTypeID    (CFUUIDGetConstantUUIDWithBytes(NULL,	\
																 0x2F,0x65,0x22,0xE9,0x3E,0x66,0x11,0xD5,\
																 0x80,0xA7,0x00,0x30,0x65,0xB3,0x00,0xBC))
// 2F6522E9-3E66-11D5-80A7-003065B300BC

// Contextual Menu Plugins must implement this Contexual Menu Plugin Interface.
#define kContextualMenuInterfaceID ( CFUUIDGetConstantUUIDWithBytes( NULL,	\
																	 0x32,0x99,0x7B,0x62,0x3E,0x66,0x11,0xD5,\
																	 0xBE,0xAB,0x00,0x30,0x65,0xB3,0x00,0xBC))
// 32997B62-3E66-11D5-BEAB-003065B300BC

#define CM_IUNKNOWN_C_GUTS \
	void *_reserved; \
	SInt32 (*QueryInterface)(void *thisPointer, CFUUIDBytes iid, void ** ppv); \
	UInt32 (*AddRef)(void *thisPointer); \
	UInt32 (*Release)(void *thisPointer)

// The function table for the interface.
struct ContextualMenuInterfaceStruct
{
	CM_IUNKNOWN_C_GUTS;
	OSStatus ( *ExamineContext )(void*               thisInstance,
							  const AEDesc*       inContext,
							  AEDescList*         outCommandPairs );
	OSStatus ( *HandleSelection )(void*               thisInstance,
							   AEDesc*             inContext,
							   SInt32              inCommandID );
	void ( *PostMenuCleanup )(void*               thisInstance );
};
typedef struct ContextualMenuInterfaceStruct ContextualMenuInterfaceStruct;

#define malloc(s) NewPtr(s)
#define free(p) DisposePtr((Ptr) p)

#include <stdio.h>
#include <string.h>

#else
#include <CoreFoundation/CoreFoundation.h>
#include <CoreFoundation/CFPlugInCOM.h>
#include <ApplicationServices/ApplicationServices.h>
#include <Carbon/Carbon.h>
#include <CoreServices/CoreServices.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>

#endif

// =============================================================================
//	typedefs, structs, consts, enums, etc.
// =============================================================================

/* 4DC35DF2-59B2-11D9-AA61-000A958FC8AC */
#define kCleanupPluginFactoryID	( CFUUIDGetConstantUUIDWithBytes( NULL, \
									0x4D,0xC3,0x5D,0xF2,0x59,0xB2,0x11,0xD9, \
									0xAA,0x61,0x00,0x0A,0x95,0x8F,0xC8,0xAC ) )

// The layout for an instance of CleanupPluginType.
typedef struct CleanupPluginType
{
	ContextualMenuInterfaceStruct	*cmInterface;
	CFUUIDRef						factoryID;
	UInt32							refCount;
} CleanupPluginType;

// =============================================================================
//	local function prototypes
// =============================================================================

// =============================================================================
//	interface function prototypes
// =============================================================================
static HRESULT CleanupPluginQueryInterface(void* thisInstance,REFIID iid,LPVOID* ppv);
static ULONG CleanupPluginAddRef(void *thisInstance);
static ULONG CleanupPluginRelease(void *thisInstance);

static OSStatus CleanupPluginExamineContext(void* thisInstance,const AEDesc* inContext,AEDescList* outCommandPairs);
static OSStatus CleanupPluginHandleSelection(void* thisInstance,AEDesc* inContext,SInt32 inCommandID);
static void CleanupPluginPostMenuCleanup( void *thisInstance );

// =============================================================================
//	static (local) global variables
// =============================================================================

static SInt32	gNumCommandIDs = 0;

// =============================================================================
//	local function implementations
// =============================================================================

// Note: If not NULL the results have to be DisposePtr'ed.
static char* Copy_CFStringRefToCString(const CFStringRef pCFStringRef)
{
    char* results = NULL;
	if (NULL != pCFStringRef)
	{
		CFIndex length = sizeof(UniChar) * CFStringGetLength(pCFStringRef) + 1;
		results = (char*) NewPtrClear(length);
		if (!CFStringGetCString(pCFStringRef,results,length,kCFStringEncodingASCII))
		{
			if (!CFStringGetCString(pCFStringRef,results,length,kCFStringEncodingUTF8))
			{
				DisposePtr(results);
				results = NULL;
			}
		}
	}
	return results;
}
// -----------------------------------------------------------------------------
//	DeallocCleanupPluginType
// -----------------------------------------------------------------------------
//	Utility function that deallocates the instance when
//	the refCount goes to zero.
//
static void DeallocCleanupPluginType( CleanupPluginType* thisInstance )
{
	CFUUIDRef	theFactoryID = thisInstance->factoryID;

	free( thisInstance );
	if ( theFactoryID )
	{
		CFPlugInRemoveInstanceForFactory( theFactoryID );
		CFRelease( theFactoryID );
	}
}

// -----------------------------------------------------------------------------
//	testInterfaceFtbl	definition
// -----------------------------------------------------------------------------
//	The TestInterface function table.
//
static ContextualMenuInterfaceStruct testInterfaceFtbl =
{
	// Required padding for COM
	NULL,

	// These three are the required COM functions
	CleanupPluginQueryInterface,
	CleanupPluginAddRef,
	CleanupPluginRelease,

	// Interface implementation
	CleanupPluginExamineContext,
	CleanupPluginHandleSelection,
	CleanupPluginPostMenuCleanup
};

#if TARGET_RT_MAC_CFM

// -----------------------------------------------------------------------------
//	MachO<->CFM Kludge
// -----------------------------------------------------------------------------
//	This function allocates a block of CFM glue code which contains the instructions to call CFM routines
//
UInt32 _template[6] = {
	0x3D800000,	// 	lis		r12,0x0000			// load ms-word TVector -> r12
	0x618C0000,	//	ori		r12,r12,0x0000		// OR in ls-word TVector -> r12
	0x800C0000,	//	lwz		r0,0x0000(r12)		// get new PC from TVector
	0x804C0004,	//	lwz		RTOC,0x004(r12)		// get new RTOC from TVector
	0x7C0903A6,	//	mtctr	r0					// new PC -> control register
	0x4E800420	//	bctr						// branch to control register
};

static void *MachOFunctionPointerForCFMFunctionPointer( void *cfmfp )
{
    UInt32	*mfp = (UInt32*) NewPtr( sizeof(_template) );		//	Must later dispose of allocated memory
    mfp[0] = _template[0] | ((UInt32)cfmfp >> 16);
    mfp[1] = _template[1] | ((UInt32)cfmfp & 0xFFFF);
    mfp[2] = _template[2];
    mfp[3] = _template[3];
    mfp[4] = _template[4];
    mfp[5] = _template[5];
    MakeDataExecutable( mfp, sizeof(_template) );

    return (mfp);
}

#define GLUE(x) MachOFunctionPointerForCFMFunctionPointer(x)

extern OSErr MyCFragInitFunction(const CFragInitBlock * initBlock);
OSErr MyCFragInitFunction(const CFragInitBlock * initBlock)
{
	(initBlock);

	(void*) testInterfaceFtbl.QueryInterface = GLUE(CleanupPluginQueryInterface);
	(void*) testInterfaceFtbl.AddRef = GLUE(CleanupPluginAddRef);
	(void*) testInterfaceFtbl.Release = GLUE(CleanupPluginRelease);
	(void*) testInterfaceFtbl.ExamineContext = GLUE(CleanupPluginExamineContext);
	(void*) testInterfaceFtbl.HandleSelection = GLUE(CleanupPluginHandleSelection);
	(void*) testInterfaceFtbl.PostMenuCleanup = GLUE(CleanupPluginPostMenuCleanup);

	return noErr;
}

extern void MyCFragTermProcedure(void);
void MyCFragTermProcedure(void)
{
	if (NULL != testInterfaceFtbl.QueryInterface)
		DisposePtr((Ptr) testInterfaceFtbl.QueryInterface);
	if (NULL != testInterfaceFtbl.AddRef)
		DisposePtr((Ptr) testInterfaceFtbl.AddRef);
	if (NULL != testInterfaceFtbl.Release)
		DisposePtr((Ptr) testInterfaceFtbl.Release);
	if (NULL != testInterfaceFtbl.ExamineContext)
		DisposePtr((Ptr) testInterfaceFtbl.ExamineContext);
	if (NULL != testInterfaceFtbl.HandleSelection)
		DisposePtr((Ptr) testInterfaceFtbl.HandleSelection);
	if (NULL != testInterfaceFtbl.PostMenuCleanup)
		DisposePtr((Ptr) testInterfaceFtbl.PostMenuCleanup);
}

#endif TARGET_RT_MAC_CFM

// -----------------------------------------------------------------------------
//	CLEANUP FILE
// -----------------------------------------------------------------------------
static int cleanup_file(char* filename, int depth)
{
	DIR				*dp;
	struct	dirent	*entry;
	struct	stat	statbuf;
	char			*trashdir,
					*tofile;
	int				trashlen;
	
	/* file must exist, be a directory and readable */
	if ((dp = opendir(filename)) == NULL) {
		printf("Error: file \"%s\" does not exist or is not readable !\n", filename);
		return 2;
	}

	/* ~/.Trash must exist */
	trashlen = strlen(getenv("HOME")) + 1;
	trashdir = malloc(trashlen + strlen("/.Trash") + 1);
	strncpy(trashdir, getenv("HOME"), trashlen);
	strncat(trashdir, "/.Trash", 7);
	if (lstat(trashdir, &statbuf) != 0) {
		printf("Error: %s is no directory or does not exist !\n", trashdir);
		return 3;
	}

	chdir(filename);
	/* read contents of directory */
	while ((entry = readdir(dp)) != NULL) {
		lstat(entry->d_name, &statbuf);
		if (S_ISDIR(statbuf.st_mode)) {
			if ( strcmp(".",entry->d_name) == 0 ||
			     strcmp("..",entry->d_name) == 0
			   )
				continue;
			printf("%*s found subdirectory: %s\n", depth, " ", entry->d_name);
			cleanup_file(entry->d_name,depth+4);
		} else if ( strcmp(".DS_Store",entry->d_name) == 0 ||
			    strncmp("._",entry->d_name,2) == 0
			  ) {
			/* move file to Trash */
			tofile = malloc( strlen(trashdir) + 2 + strlen(entry->d_name) + 1);
			strcpy(tofile, trashdir);
			strcat(tofile, "/");
			strcat(tofile, entry->d_name);
			if (rename(entry->d_name, tofile) != 0) {
				/* file can't be moved, lets copy it */
				char	block[1024];
				int		in, out;
				int		nread;
				
				unlink(tofile); /* delete destination first */
				in  = open(entry->d_name, O_RDONLY);
				out = open(tofile, O_WRONLY|O_CREAT);
				if ((in > 0) && (out > 0)) {
					while ((nread=read(in,block,sizeof(block))) > 0)
						write(out,block,nread);
					close(in);
					close(out);
					if (unlink(entry->d_name) != 0) {
						printf("Error: \"%s\" can't be removed !\n", entry->d_name);
					} else {
						printf("%*s REMOVED \"%s\", COPIED to \"%s\" !\n", depth, " ", entry->d_name, tofile);
					}
				} else {
					printf("Error: \"%s\" can't be copied !\n", entry->d_name);
				}				
			} else {
				printf("%*s MOVED \"%s\" to \"%s\"\n", depth, " ", entry->d_name, tofile);
			}
		} else {
			printf("%*s regular file: %s\n", depth, " ", entry->d_name);
		}
	}
	chdir("..");
	closedir(dp);
	
	return 0;
}

// -----------------------------------------------------------------------------
//	AllocCleanupPluginType
// -----------------------------------------------------------------------------
//	Utility function that allocates a new instance.
//
static CleanupPluginType* AllocCleanupPluginType(CFUUIDRef		inFactoryID )
{
	// Allocate memory for the new instance.
	CleanupPluginType *theNewInstance;

	theNewInstance = (CleanupPluginType*) malloc(sizeof( CleanupPluginType ) );

	// Point to the function table
	theNewInstance->cmInterface = &testInterfaceFtbl;

	// Retain and keep an open instance refcount<
	// for each factory.
	theNewInstance->factoryID = CFRetain( inFactoryID );
	CFPlugInAddInstanceForFactory( inFactoryID );

	// This function returns the IUnknown interface
	// so set the refCount to one.
	theNewInstance->refCount = 1;
	return theNewInstance;
}

// -----------------------------------------------------------------------------
//	CleanupPluginFactory
// -----------------------------------------------------------------------------
//	Implementation of the factory function for this type.
//
extern void* CleanupPluginFactory(CFAllocatorRef allocator,CFUUIDRef typeID );
void* CleanupPluginFactory(CFAllocatorRef allocator,CFUUIDRef typeID )
{
#pragma unused (allocator)

	// If correct type is being requested, allocate an
	// instance of TestType and return the IUnknown interface.
	if ( CFEqual( typeID, kContextualMenuTypeID ) )
	{
		CleanupPluginType *result;
		result = AllocCleanupPluginType(kCleanupPluginFactoryID);
		return result;
	}
	else
	{
		// If the requested type is incorrect, return NULL.
		return NULL;
	}
}

// -----------------------------------------------------------------------------
//	AddCommandToAEDescList
// -----------------------------------------------------------------------------
static OSStatus AddCommandToAEDescList(CFStringRef	inCommandCFStringRef,
									   SInt32		inCommandID,
									   AEDescList*	ioCommandList)
{
	OSStatus anErr = noErr;
	AERecord theCommandRecord = { typeNull, NULL };
	CFStringRef tCFStringRef = CFStringCreateWithFormat(kCFAllocatorDefault,NULL,CFSTR("%@"),inCommandCFStringRef);
	CFIndex length = CFStringGetLength(tCFStringRef);
    const UniChar* dataPtr = CFStringGetCharactersPtr(tCFStringRef);
	const UniChar* tempPtr = nil;

	//	printf("CleanupPlugin->AddCommandToAEDescList: Trying to add an item.\n" );

    if (dataPtr == NULL)
	{
		tempPtr = (UniChar*) NewPtr(length * sizeof(UniChar));
		if (nil == tempPtr)
			goto AddCommandToAEDescList_fail;

		CFStringGetCharacters(tCFStringRef, CFRangeMake(0,length), (UniChar*) tempPtr);
		dataPtr = tempPtr;
	}
	if (nil == dataPtr)
		goto AddCommandToAEDescList_fail;

	// create an apple event record for our command
	anErr = AECreateList( NULL, 0, true, &theCommandRecord );
	require_noerr( anErr, AddCommandToAEDescList_fail );

	// stick the command text into the aerecord
	anErr = AEPutKeyPtr( &theCommandRecord, keyAEName, typeUnicodeText, dataPtr, length * sizeof(UniChar));
	require_noerr( anErr, AddCommandToAEDescList_fail );

	// stick the command ID into the AERecord
	anErr = AEPutKeyPtr( &theCommandRecord, keyContextualMenuCommandID,
					  typeLongInteger, &inCommandID, sizeof( inCommandID ) );
	require_noerr( anErr, AddCommandToAEDescList_fail );

	// stick this record into the list of commands that we are
	// passing back to the CMM
	anErr = AEPutDesc(ioCommandList, 			// the list we're putting our command into
				   0, 						// stick this command onto the end of our list
				   &theCommandRecord );	// the command I'm putting into the list

AddCommandToAEDescList_fail:
		// clean up after ourself; dispose of the AERecord
		AEDisposeDesc( &theCommandRecord );

	if (nil != tempPtr)
		DisposePtr((Ptr) tempPtr);

    return anErr;

} // AddCommandToAEDescList

// -----------------------------------------------------------------------------
//	CreateFileSubmenu
// -----------------------------------------------------------------------------
static OSStatus CreateFileSubmenu(const AEDesc* inContext,AEDescList* ioCommandList)
{
	OSStatus	anErr = noErr;
	AEDescList	theSubmenuCommands = { typeNull, NULL };
	AERecord	theSupercommand = { typeNull, NULL };
	Str255		theSupercommandText = "\pSubmenu Here";

	// set up the AEDescList
	anErr = AECreateList( NULL, 0, false, &theSubmenuCommands );
	require_noerr( anErr, CreateFileSubmenu_Complete_fail );

	{
		long	numItems;
		OSErr   anErr = AECountItems(inContext,&numItems);
		if (noErr == anErr) {
			sprintf((Ptr) &theSupercommandText[1],"CLEANUP THESE %ld OBJECTS...",numItems);
			theSupercommandText[0] = strlen((Ptr) &theSupercommandText[1]);
		}
	}

	// now, we need to create the supercommand which will "own" the
	// subcommands.  The supercommand lives in the root command list.
	// this looks very much like the AddCommandToAEDescList function,
	// except that instead of putting a command ID in the record,
	// we put in the subcommand list.

	// create an apple event record for our supercommand
	anErr = AECreateList( NULL, 0, true, &theSupercommand );
	require_noerr( anErr, CreateFileSubmenu_fail );

	// stick the command text into the aerecord
	anErr = AEPutKeyPtr(&theSupercommand, keyAEName, typeChar,&theSupercommandText[1], theSupercommandText[0]);
	require_noerr( anErr, CreateFileSubmenu_fail );

	// stick the subcommands into into the AERecord
	anErr = AEPutKeyDesc(&theSupercommand, keyContextualMenuSubmenu,&theSubmenuCommands);
	require_noerr( anErr, CreateFileSubmenu_fail );

	// stick the supercommand into the list of commands that we are
	// passing back to the CMM
	anErr = AEPutDesc(
				   ioCommandList,		// the list we're putting our command into
				   0,					// stick this command onto the end of our list
				   &theSupercommand);	// the command I'm putting into the list

	// clean up after ourself
CreateFileSubmenu_fail:
		AEDisposeDesc(&theSubmenuCommands);

	AEDisposeDesc(&theSupercommand);

CreateFileSubmenu_Complete_fail:
		return anErr;

} // CreateFileSubmenu

// -----------------------------------------------------------------------------
//	HandleFileSubmenu
// -----------------------------------------------------------------------------
static OSStatus HandleFileSubmenu(const AEDesc* inContext,const SInt32 inCommandID)
{
	long	index,numItems;
	OSStatus	anErr = noErr;

	// printf("CleanupPlugin->HandleFileSubmenu(inContext: 0x%lx, inCommandID: 0x%lx)\n",(SInt32) inContext,(SInt32) inCommandID );

	anErr = AECountItems(inContext,&numItems);
	if (noErr == anErr) {
		for (index = 1;index <= numItems;index++) {
			AEKeyword	theAEKeyword;
			AEDesc		theAEDesc;

			anErr = AEGetNthDesc(inContext,index,typeAlias,&theAEKeyword,&theAEDesc);
			if (noErr != anErr) continue;

			if (theAEDesc.descriptorType == typeAlias) {
				FSRef tFSRef;
				Size dataSize = AEGetDescDataSize(&theAEDesc);
				AliasHandle tAliasHdl = (AliasHandle) NewHandle(dataSize);

				if (nil != tAliasHdl) {
					Boolean wasChanged;

					anErr = AEGetDescData(&theAEDesc,*tAliasHdl,dataSize);
					if (noErr == anErr) {
						anErr = FSResolveAlias(NULL,tAliasHdl,&tFSRef,&wasChanged);
						if (noErr == anErr) {
							/* see /Developer/Documentation/CoreFoundation/Reference/CFURLRef/Reference/function_group_2.html */
							CFURLRef tCFURLRef = CFURLCreateFromFSRef(kCFAllocatorDefault, &tFSRef);
							CFStringRef nameCFStringRef = CFURLCopyFileSystemPath(tCFURLRef,kCFURLPOSIXPathStyle);

							CFRelease(tCFURLRef);	// don't need this anymore

							if (NULL != nameCFStringRef) {
								char* tPtr = Copy_CFStringRefToCString(nameCFStringRef);
								if (nil != tPtr) {
									printf("SELECTED FILE: \"%s\"\n",tPtr);
									cleanup_file(tPtr, 0);
									DisposePtr(tPtr);
								}
								CFRelease(nameCFStringRef);	// don't need this anymore
							}
						}
					}
					DisposeHandle((Handle) tAliasHdl);
				}
			}
			AEDisposeDesc(&theAEDesc);
		}
	}

	return anErr;
} // HandleFileSubmenu

// -----------------------------------------------------------------------------
//	CreateSampleSubmenu
// -----------------------------------------------------------------------------
static OSStatus CreateSampleSubmenu(AEDescList*		ioCommandList)
{
	OSStatus	anErr = noErr;

	AEDescList	theSubmenuCommands = { typeNull, NULL };
	AERecord	theSupercommand = { typeNull, NULL };
	Str255		theSupercommandText = "\pSubmenu Here";

	// the first thing we should do is create an AEDescList of subcommands

	// set up the AEDescList
	anErr = AECreateList( NULL, 0, false, &theSubmenuCommands );
	require_noerr( anErr, CreateSampleSubmenu_Complete_fail );

	// stick some commands in this subcommand list
	anErr = AddCommandToAEDescList( CFSTR("Subcommand 1"), 0x00020001,
								 &theSubmenuCommands );
	require_noerr( anErr, CreateSampleSubmenu_CreateDesc_fail );

	// another
	anErr = AddCommandToAEDescList( CFSTR("Another Subcommand"), 0x00020002,
								 &theSubmenuCommands );
	require_noerr( anErr, CreateSampleSubmenu_fail );

	// yet another
	anErr = AddCommandToAEDescList( CFSTR("Last One"), 0x00020003,
								 &theSubmenuCommands);
	require_noerr( anErr, CreateSampleSubmenu_fail );

	// now, we need to create the supercommand which will "own" the
	// subcommands.  The supercommand lives in the root command list.
	// this looks very much like the AddCommandToAEDescList function,
	// except that instead of putting a command ID in the record,
	// we put in the subcommand list.

	// create an apple event record for our supercommand
	anErr = AECreateList( NULL, 0, true, &theSupercommand );
	require_noerr( anErr, CreateSampleSubmenu_fail );

	// stick the command text into the aerecord
	anErr = AEPutKeyPtr(&theSupercommand, keyAEName, typeChar,
					 &theSupercommandText[1], theSupercommandText[0]);
	require_noerr( anErr, CreateSampleSubmenu_fail );

	// stick the subcommands into into the AERecord
	anErr = AEPutKeyDesc(&theSupercommand, keyContextualMenuSubmenu,
					  &theSubmenuCommands);
	require_noerr( anErr, CreateSampleSubmenu_fail );

	// stick the supercommand into the list of commands that we are
	// passing back to the CMM
	anErr = AEPutDesc(
				   ioCommandList,		// the list we're putting our command into
				   0,					// stick this command onto the end of our list
				   &theSupercommand);	// the command I'm putting into the list

	// clean up after ourself
CreateSampleSubmenu_fail:
		AEDisposeDesc(&theSubmenuCommands);

CreateSampleSubmenu_CreateDesc_fail:
		AEDisposeDesc(&theSupercommand);

CreateSampleSubmenu_Complete_fail:
		return anErr;

} // CreateSampleSubmenu

// -----------------------------------------------------------------------------
//	HandleSampleSubmenu
// -----------------------------------------------------------------------------
static OSStatus HandleSampleSubmenu(const AEDesc* inContext,const SInt32 inCommandID)
{
	OSStatus	anErr = noErr;

#pragma unused (inContext)

	switch (inCommandID)
	{
		case 0x00020001: {
			printf("\"Subcommand 1\"\n");
			cleanup_file("test", 0);
			break;
		}
		case 0x00020002:
			printf("\"Another Subcommand\"\n"); break;
		case 0x00020003:
			printf("\"Last One\"\n"); break;
		default:
			if (0x0002 == (inCommandID >> 16))
				printf("HandleSampleSubmenu:(inCommandID: %.8lX)\n",inCommandID); break;
	}

	return anErr;
} // HandleSampleSubmenu

// =============================================================================
//	interface function implementations
// =============================================================================

// -----------------------------------------------------------------------------
//	CleanupPluginQueryInterface
// -----------------------------------------------------------------------------
//	Implementation of the IUnknown QueryInterface function.
//
static HRESULT CleanupPluginQueryInterface(void* thisInstance,REFIID iid,LPVOID* ppv)
{
	// Create a CoreFoundation UUIDRef for the requested interface.
	CFUUIDRef	interfaceID = CFUUIDCreateFromUUIDBytes( NULL, iid );

	// Test the requested ID against the valid interfaces.
	if ( CFEqual( interfaceID, kContextualMenuInterfaceID ) )
	{
		// If the TestInterface was requested, bump the ref count,
		// set the ppv parameter equal to the instance, and
		// return good status.
  //		((CleanupPluginType*) thisInstance )->cmInterface->AddRef(thisInstance );
		CleanupPluginAddRef(thisInstance);

		*ppv = thisInstance;
		CFRelease( interfaceID );
		return S_OK;
	}
	else if ( CFEqual( interfaceID, IUnknownUUID ) )
	{
		// If the IUnknown interface was requested, same as above.
  //		( ( CleanupPluginType* ) thisInstance )->cmInterface->AddRef(thisInstance );
		CleanupPluginAddRef(thisInstance);

		*ppv = thisInstance;
		CFRelease( interfaceID );
		return S_OK;
	}
	else
	{
		// Requested interface unknown, bail with error.
		*ppv = NULL;
		CFRelease( interfaceID );
		return E_NOINTERFACE;
	}
}

// -----------------------------------------------------------------------------
//	CleanupPluginAddRef
// -----------------------------------------------------------------------------
//	Implementation of reference counting for this type. Whenever an interface
//	is requested, bump the refCount for the instance. NOTE: returning the
//	refcount is a convention but is not required so don't rely on it.
//
static ULONG CleanupPluginAddRef( void *thisInstance )
{
	( ( CleanupPluginType* ) thisInstance )->refCount += 1;
	return ( ( CleanupPluginType* ) thisInstance)->refCount;
}

// -----------------------------------------------------------------------------
// CleanupPluginRelease
// -----------------------------------------------------------------------------
//	When an interface is released, decrement the refCount.
//	If the refCount goes to zero, deallocate the instance.
//
static ULONG CleanupPluginRelease( void *thisInstance )
{
	((CleanupPluginType*) thisInstance )->refCount -= 1;
	if (((CleanupPluginType*) thisInstance)->refCount == 0)
	{
		DeallocCleanupPluginType((CleanupPluginType*) thisInstance);
		return 0;
	}
	else
	{
		return ((CleanupPluginType*) thisInstance)->refCount;
	}
}

// -----------------------------------------------------------------------------
//	CleanupPluginExamineContext
// -----------------------------------------------------------------------------
//	The implementation of the ExamineContext test interface function.
//

static OSStatus CleanupPluginExamineContext(	void*				thisInstance,
											  const AEDesc*		inContext,
											  AEDescList*			outCommandPairs)
{
//	CFStringRef bundleCFStringRef = CFSTR("com.korseby.CleanupPlugin");
//	CFBundleRef myCFBundleRef = CFBundleGetBundleWithIdentifier(bundleCFStringRef);

	// Sequence the command ids
	gNumCommandIDs = 0;

/*	printf("CleanupPlugin->CleanupPluginExamineContext(instance: 0x%x, inContext: 0x%x, outCommandPairs: 0x%x)\n",
		( unsigned ) thisInstance,
		( const unsigned ) inContext,
		( unsigned ) outCommandPairs );
*/
	// this is a quick sample that looks for text in the context descriptor

	// make sure the descriptor isn't null
	if ( inContext != NULL )
	{
		AEDesc theAEDesc = { typeNull, NULL };

		if (inContext->descriptorType == typeAEList) {
			long numItems;
			AECountItems(inContext,&numItems);
			if (numItems == 1) {
				OSErr anErr;
				Boolean wasChanged;
				AEKeyword	theAEKeyword;
				anErr = AEGetNthDesc(inContext,1,typeAlias,&theAEKeyword,&theAEDesc);
				//if (noErr != anErr) continue;

				FSRef tFSRef;
				Size dataSize = AEGetDescDataSize(&theAEDesc);
				AliasHandle tAliasHdl = (AliasHandle) NewHandle(dataSize);
				anErr = AEGetDescData(&theAEDesc,*tAliasHdl,dataSize);
				if (noErr == anErr) {
					anErr = FSResolveAlias(NULL,tAliasHdl,&tFSRef,&wasChanged);
					if (noErr == anErr) {
						CFURLRef	tCFURLRef = CFURLCreateFromFSRef(kCFAllocatorDefault, &tFSRef);
						CFStringRef nameCFStringRef = CFURLCopyLastPathComponent(tCFURLRef);
						nameCFStringRef = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, CFSTR("CLEANUP \"%@\""), nameCFStringRef);
						CFRelease(tCFURLRef);
						AddCommandToAEDescList (nameCFStringRef, ++gNumCommandIDs, outCommandPairs);
						CFRelease(nameCFStringRef);
					}
				}
				DisposeHandle((Handle) tAliasHdl);
				AEDisposeDesc( &theAEDesc );
			} else {
				CreateFileSubmenu(inContext,outCommandPairs);
			}
		} else if (inContext->descriptorType == typeAlias)
		{
			if ( AECoerceDesc( inContext, typeAEList, &theAEDesc ) == noErr )
			{
				CreateFileSubmenu(&theAEDesc,outCommandPairs);
				AEDisposeDesc( &theAEDesc );
			}
			else
			{
				// add a text only command to our command list
				AddCommandToAEDescList( CFSTR("Can't Coerce alias to list."), ++gNumCommandIDs,outCommandPairs );
			}
		}
		else
		{
			// try to get text out of the context descriptor; make sure to
			// coerce it, cuz the app may have passed an object specifier or
			// styled text, etc.
			if ( AECoerceDesc( inContext, typeChar, &theAEDesc ) == noErr )
			{
				// add a text only command to our command list
				AddCommandToAEDescList( CFSTR("We got text!"), ++gNumCommandIDs,outCommandPairs );
				AEDisposeDesc( &theAEDesc );
			}
			else
			{
				// printf("CleanupPluginExamineContext: Unable to coerce to text.\n" );
				// add a text only command to our command list
				AddCommandToAEDescList( CFSTR("Can't Coerce."), ++gNumCommandIDs,outCommandPairs );
			}
		}

		AddCommandToAEDescList( CFSTR("-"), ++gNumCommandIDs,outCommandPairs );
	}
	else
	{
		// printf("CleanupPluginExamineContext: Hey! What's up with the NULL descriptor?\n" );
		// we have a null descriptor
		AddCommandToAEDescList( CFSTR("NULL Descriptor"), ++gNumCommandIDs,outCommandPairs );
	}

	return noErr;
}

// -----------------------------------------------------------------------------
//	HandleSelection
// -----------------------------------------------------------------------------
//	The implementation of the HandleSelection test interface function.
//
static OSStatus CleanupPluginHandleSelection(void*				thisInstance,
											  AEDesc*				inContext,
											  SInt32				inCommandID )
{
	// Sequence the command ids
	gNumCommandIDs = 0;

	// this is a quick sample that looks for text in the context descriptor

	// make sure the descriptor isn't null
	if ( inContext != NULL )
	{
		AEDesc theAEDesc = { typeNull, NULL };

		if (inContext->descriptorType == typeAEList)
			HandleFileSubmenu(inContext,inCommandID);
		else if (inContext->descriptorType == typeAlias)
		{
			if ( AECoerceDesc( inContext, typeAEList, &theAEDesc ) == noErr )
			{
				HandleFileSubmenu(&theAEDesc,inCommandID);
				AEDisposeDesc( &theAEDesc );
			}
			else
			{
				printf("CleanupPluginHandleSelection: Unable to coerce to list.\n" );
			}
		}
		else
		{
			// try to get text out of the context descriptor; make sure to
   // coerce it, cuz the app may have passed an object specifier or
   // styled text, etc.
			if ( AECoerceDesc( inContext, typeChar, &theAEDesc ) == noErr )
			{
				// add a text only command to our command list
	//				AddCommandToAEDescList( CFSTR("We got text!"), ++gNumCommandIDs,outCommandPairs );
				printf("CleanupPluginHandleSelection: ..............\n" );
				++gNumCommandIDs;
				AEDisposeDesc( &theAEDesc );
			}
			else
			{
				printf("CleanupPluginHandleSelection: Unable to coerce to text.\n" );
				// add a text only command to our command list
	//				AddCommandToAEDescList( CFSTR("Can't Coerce."), ++gNumCommandIDs,outCommandPairs );
				++gNumCommandIDs;
			}
		}

		// handle the submenu for our plugin
//		HandleSampleSubmenu(inContext,inCommandID);
	}
	else
	{
		printf("CleanupPluginHandleSelection: Hey! What's up with the NULL descriptor?\n");
		// we have a null descriptor
  //		AddCommandToAEDescList( CFSTR("NULL Descriptor"), ++gNumCommandIDs,outCommandPairs );
		++gNumCommandIDs;
	}

	return noErr;
}	// CleanupPluginHandleSelection

// -----------------------------------------------------------------------------
//	PostMenuCleanup
// -----------------------------------------------------------------------------
static void CleanupPluginPostMenuCleanup( void *thisInstance )
{
	
}
